// Released under the MIT licence.
// See LICENCE.txt for details.

#include "Backend.h"
#include "../../Misc.h"

#include <csignal>
#include <stdint.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdbool.h>
#include <sys/audioio.h>
#include <pthread.h>

static const int samplesPerGo = 0x200;
static const int parkSignal = SIGUSR1;
static pthread_t mainThread;
static int sndfp;
static FILE* sndfile;
static void (*parent_callback)(long *stream, size_t frames_total);
static bool audioDone = false;

static void soundSetup() {
}

static void* soundcheckthread(void*) {
    printf("Entered sound thread.\n");
    unsigned long for_as_long_as_it_seems_we_should = 0;
    bool firstTime = true;
    int counter = 0;
    while(1) {
        long mix[samplesPerGo * 2];
        memset(mix, 0, samplesPerGo * 2 * sizeof(long));

        parent_callback(mix, samplesPerGo);

        for_as_long_as_it_seems_we_should = ((float)(samplesPerGo) / (float)8000) * 1000000;

        audio_info_t audioInfo;
        ioctl(sndfp, AUDIO_GETINFO, &audioInfo);
        //printf("%d\n", audioInfo.play.error);

        if(counter > 4 && audioInfo.play.error == 0) {
            usleep(for_as_long_as_it_seems_we_should);
        } else {
            printf("ae\n");
            audioInfo.play.error = 0;
            ioctl(sndfp, AUDIO_SETINFO, &audioInfo);
        }

        for(int i = 0; i < samplesPerGo * 2; i++) {
            //if(i % 2 == 0) {
                short sample = mix[i];
                if(mix[i] > 0x7FFF) {
                    sample = 0x7FFF;
                }
                if(mix[i] < -0x7FFF) {
                    sample = -0x7FFF;
                }
                fputc((sample >> 8) & 0xFF, sndfile);
                fputc(sample & 0xFF, sndfile);
            //}
        }

        counter += 1;
        firstTime = false;
    }
}

static void threadPark(int signo) {
    while(audioDone == false);
}

unsigned long SoftwareMixerBackend_Init(void (*callback)(long *stream, size_t frames_total)) {
    parent_callback = callback;

    if(sndfp == -1) {
        Backend_PrintError("Failed to open audio device.\n");
        return 0;
    }

    sndfile = fopen("/dev/audio", "wb");
    sndfp = fileno(sndfile);
    audio_info_t audioInfo;
    ioctl(sndfp, AUDIO_GETINFO, &audioInfo);
    audio_prinfo_t info = audioInfo.play;

    info.sample_rate = 8000;
    info.channels = 2;
    info.precision = 8;
    info.encoding = AUDIO_ENCODING_LINEAR;
    info.port = AUDIO_SPEAKER | AUDIO_HEADPHONE;
    info.gain = 128;

    audioInfo.play = info;
    ioctl(sndfp, AUDIO_SETINFO, &audioInfo);
    //close(sndfp);

    if(signal(parkSignal, threadPark) == SIG_ERR) {
        Backend_PrintError("Failed to register thread park signal handler for audio system.\n");
        return 0;
    }

    mainThread = pthread_self();

    pthread_t thread;
    pthread_attr_t thread_attr;
    pthread_attr_init(&thread_attr);

    pthread_create(&thread, &thread_attr, soundcheckthread, (void*)NULL);

    return 8000; // return the sampling rate
}

void SoftwareMixerBackend_Deinit(void) {
    close(sndfp);
}

bool SoftwareMixerBackend_Start(void) {
    return true;
}

void SoftwareMixerBackend_LockMixerMutex(void) {
}

void SoftwareMixerBackend_UnlockMixerMutex(void) {
}

void SoftwareMixerBackend_LockOrganyaMutex(void) {
}

void SoftwareMixerBackend_UnlockOrganyaMutex(void) {
}
